import numpy as np
import matplotlib.pyplot as plt
from sklearn.decomposition import SparseCoder

import src.archive.dictionary as dict
import src.utils as utils
import src.factorization as fact
import argparse
from pathlib import Path


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Learning a dictionary for MEG with methods of optimal directions")
    parser.add_argument("--data", type=str, help="Path to MEG matrix in .npy format")
    parser.add_argument("--k", type=int, help="Sparsity by column on the sparsity")
    parser.add_argument("--random_permute_MEG", action="store_true")
    parser.add_argument("--coder", choices=["omp", "lars"], help="Method for SparseCoder from scikit-learn")
    parser.add_argument("--code_init", choices=["random", "eye"])
    parser.add_argument("--iter", type=int, help="Number of iteration in the dictionary learning algorithm")
    parser.add_argument("--global_optim", action="store_true")
    parser.add_argument("--lr", type=float)
    parser.add_argument("--grad_iter", type=int)
    parser.add_argument("--grad_optimizer", choices=["SGD", "SGDMomentum", "Adam"])
    parser.add_argument("--results_path", type=str, help="Directory path for experiments results")
    parser.add_argument("--rewrite_results", action="store_true")
    args = parser.parse_args()

    # Dump
    name = f"{'random_permute_MEG' if args.random_permute_MEG else 'no_permute_MEG'}-" \
           f"{args.code_init}_code_init-{'global_optim' if args.global_optim else 'no_global_optim'}-" \
           f"lr_{args.lr}-grad_iter_{args.grad_iter}-{args.grad_optimizer}"
    dir_path = Path(args.results_path)
    dir_path.mkdir(exist_ok=True, parents=True)
    results = utils.Stats(dir_path / f"{name}.pkl",
                          ["k", "coder", "random_permute_MEG", "code_init", "global_optim",
                           "lr", "grad_iter", "grad_optimizer", "iter", "loss"],
                          rewrite=args.rewrite_results)
    # Main
    MEG = np.load(args.data)
    print("Shape of MEG matrix:", MEG.shape)
    code, dictionary = dict.method_of_optimal_directions(
        MEG,
        stats=results,
        col_sparsity=args.k,
        num_iter=args.iter,
        coder_method=args.coder,
        random_permute_MEG=args.random_permute_MEG,
        code_init=args.code_init,
        global_optim=None,
        lr=None,
        optimize=None,
        grad_iter=None
    )
    print("Final loss before projection: ", utils.relative_error_padded_product(MEG, code, dictionary))

    # Projection
    projection = fact.project_BP_model_P_fixed(dictionary, "comb", np.identity(dictionary.shape[0]), max_depth=1, )
    norms = np.linalg.norm(projection, axis=1)
    indices = np.argwhere(norms > 0)
    dictionary[indices] = projection[indices] / norms[indices, np.newaxis]
    coder = SparseCoder(
        dictionary=dictionary,
        transform_n_nonzero_coefs=args.k,
        transform_algorithm=args.coder
    )
    code = coder.transform(utils.add_padding(MEG))
    print("Final loss after projection: ", utils.relative_error_padded_product(MEG, code, projection))

    # Plot learning graph
    plt.plot(results.stats["iter"], results.stats["loss"])
    plt.title(name)
    plt.xlabel("iteration")
    plt.ylabel("relative error")
    plt.savefig(dir_path / f"{name}.png")
    plt.close()

    # Save
    np.save(dir_path / f"code.npy", code)
    np.save(dir_path / f"dictionary.npy", dictionary)
