import numpy as np
import matplotlib.pyplot as plt

import src.archive.dictionary as dict
import src.utils as utils
import argparse
from pathlib import Path


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Learning a dictionary + sparse coding of the MEG matrix, where"
                                                 "the dictionary can be factorized into butterfly factors")
    parser.add_argument("--data", type=str, help="Path to MEG matrix in .npy format")
    parser.add_argument("--k", type=int, help="Sparsity by column on the sparsity")
    parser.add_argument("--tree", type=str, choices=["comb", "balanced", "inversed_comb"], default="balanced")
    parser.add_argument("--max_depth", default=-1, type=int)
    parser.add_argument("--perm",
                        choices=["identity", "bit-reversal", "000", "001", "010", "011", "100", "101", "110", "111",
                                 "all_8"],
                        default="identity"
                        )
    parser.add_argument("--random_permute_MEG", action="store_true")
    parser.add_argument("--coder", choices=["omp", "lars"], help="Method for SparseCoder from scikit-learn")
    parser.add_argument("--code_init", choices=["random", "eye"])
    parser.add_argument("--iter", type=int, help="Number of iteration in the dictionary learning algorithm")
    parser.add_argument("--global_optim", action="store_true")
    parser.add_argument("--lr", type=float)
    parser.add_argument("--grad_iter", type=int)
    parser.add_argument("--grad_optimizer", choices=["SGD", "SGDMomentum", "Adam"])
    parser.add_argument("--results_path", type=str, help="Directory path for experiments results")
    parser.add_argument("--rewrite_results", action="store_true")
    args = parser.parse_args()

    # Dump
    name = f"{args.perm}_perm-{args.tree}_tree-{args.coder}-{args.k}_sparse-max_depth_{args.max_depth}-" \
           f"{'random_permute_MEG' if args.random_permute_MEG else 'no_permute_MEG'}-" \
           f"{args.code_init}_code_init-{'global_optim' if args.global_optim else 'no_global_optim'}-" \
           f"lr_{args.lr}-grad_iter_{args.grad_iter}-{args.grad_optimizer}"
    dir_path = Path(args.results_path)
    dir_path.mkdir(exist_ok=True, parents=True)
    results = utils.Stats(dir_path / f"{name}.pkl",
                          ["k", "tree", "perm", "coder", "random_permute_MEG", "code_init", "global_optim",
                           "lr", "grad_iter", "grad_optimizer", "iter", "loss"],
                          rewrite=args.rewrite_results)

    # Main
    MEG = np.load(args.data)
    print("Shape of MEG matrix:", MEG.shape)
    code, *factors, p = dict.learning_with_butterfly_model(
        MEG,
        col_sparsity=args.k,
        num_iter=args.iter,
        coder_method=args.coder,
        tree=args.tree,
        perm_name=args.perm,
        stats=results,
        max_depth=args.max_depth,
        random_permute_MEG=args.random_permute_MEG,
        code_init=args.code_init,
        global_optim=args.global_optim,
        lr=args.lr,
        optimize=args.grad_optimizer,
        grad_iter=args.grad_iter
    )
    print("Final loss: ", utils.relative_error_padded_product(MEG, code, *factors, p))

    # Plot learning graph
    plt.plot(results.stats["iter"], results.stats["loss"])
    plt.title(name)
    plt.xlabel("iteration")
    plt.ylabel("relative error")
    plt.savefig(dir_path / f"{name}.png")
    plt.close()
