import numpy as np
import argparse
from pathlib import Path
import matplotlib.pyplot as plt

import src.archive.palm as palm
import src.utils as utils

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="PALM factorization of MEG matrix")
    parser.add_argument("--data", type=str, help="Path to MEG matrix in .npy format", default="data/MEG.npy")
    parser.add_argument("--A_sparsity", type=int, default=25)
    parser.add_argument("--B_sparsity", type=int, default=25)
    parser.add_argument("--random_perm", action="store_true")
    parser.add_argument("--max_iter", type=int, help="Maximum of PALM iteration", default=100)
    parser.add_argument("--BHTP", action="store_true")
    parser.add_argument("--results_path", type=str, help="Directory path for experiments results")
    parser.add_argument("--rewrite_results", action="store_true")
    parser.add_argument("--BP_projection", action="store_true")
    parser.add_argument("--cuda", action="store_true")
    args = parser.parse_args()

    MEG = np.load(args.data)
    print("Shape of MEG matrix:", MEG.shape)

    name = f"{args.A_sparsity}_A_sparsity-{args.B_sparsity}_B_sparsity-{'BHTP' if args.BHTP else 'PALM'}-{'random_perm' if args.random_perm else 'no_perm'}"
    dir_path = Path(args.results_path)
    dir_path.mkdir(exist_ok=True, parents=True)
    results = utils.Stats(dir_path / f"{name}.pkl",
                          ["A_sparsity", "B_sparsity", "random_perm", "BHTP", "iter", "loss"],
                          rewrite=args.rewrite_results)

    if args.BP_projection:
        if args.cuda:
            print("Cuda, BP factorization")
            palm_method = palm.torch_palm_with_BP_factorization_two_factors
        else:
            print("BP factorization")
            palm_method = palm.palm_with_BP_factorization_two_factors
    else:
        palm_method = palm.palm_factorization_two_factors

    evol1 = palm_method(
        MEG,
        k0=args.A_sparsity,
        k1=args.B_sparsity,
        max_iter=args.max_iter,
        flag=args.BHTP,
        random_perm=args.random_perm,
        stats=results
    )
    print("Final relative error:", evol1[-1])

    # Plot learning graph
    plt.plot(results.stats["iter"], results.stats["loss"])
    plt.title(name)
    plt.xlabel("iteration")
    plt.ylabel("relative error")
    plt.savefig(dir_path / f"{name}.png")
    plt.close()
