import argparse
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import torch.utils.data
from torch.nn.parameter import Parameter
from torchvision import datasets, transforms
from torch.optim.lr_scheduler import StepLR
import src.utils
import numpy as np
from src.archive.masked_linear import CustomizedLinear
import src.tree
from pathlib import Path
import time


def torch_best_low_rank(matrix, rank):
    u, s, vh = torch.linalg.svd(matrix)
    s = torch.sqrt(s[:rank])
    return torch.mm(u[:, range(rank)], torch.diag(s)), torch.mm(torch.diag(s), vh[range(rank)])


def torch_lifting_two_layers_factorization(left_supp, right_supp, tensor):
    assert left_supp.shape[1] == right_supp.shape[0]
    dtype = tensor.dtype
    X = torch.zeros(left_supp.shape, dtype=dtype).to(tensor.device)
    Y = torch.zeros(right_supp.shape, dtype=dtype).to(tensor.device)
    r = left_supp.shape[1]
    for t in range(r):
        rows = torch.where(left_supp[:, t])[0]
        cols = torch.where(right_supp[t, :])[0]
        submat = tensor[torch.meshgrid(rows, cols)]
        u, v = torch_best_low_rank(submat, 1)
        X[rows, t] = torch.squeeze(u)
        Y[t, cols] = torch.squeeze(v)
    return X, Y


def torch_tree_hierarchical_factorization(tensor, root):
    assert not root.is_leaf()
    X, Y = torch_lifting_two_layers_factorization(root.left.support, root.right.support, tensor)
    left_factors = [X] if root.left.is_leaf() else torch_tree_hierarchical_factorization(X, root.left)
    right_factors = [Y] if root.right.is_leaf() else torch_tree_hierarchical_factorization(Y, root.right)
    return left_factors + right_factors


def torch_product_of_factors(factors):
    product = factors[0]
    for i in range(1, len(factors)):
        product = torch.mm(product, factors[i])
    return product


def torch_project_p_model_p_fixed(tensor, root, permutation=None):
    if permutation is not None:
        factors = torch_tree_hierarchical_factorization(tensor.mm(tensor, permutation.t()), root)
        factors.append(permutation)
    else:
        factors = torch_tree_hierarchical_factorization(tensor, root)
    product = torch_product_of_factors(factors)
    return product


class NaiveButterfly(nn.Module):
    def __init__(self, n_features, permutation=None, bias=True):
        super(NaiveButterfly, self).__init__()
        num_factors = int(np.log2(n_features))
        supports = src.utils.support_DFT(num_factors)
        if permutation is not None:
            supports[-1] = supports[-1] @ permutation
        self.layers = [CustomizedLinear(torch.from_numpy(supp).to(torch.bool), bias=None)
                       for supp in supports]
        for i, layer in enumerate(self.layers):
            for param in layer.parameters():
                self.register_parameter(name=f'butterfly-{i}', param=param)
        if bias:
            self.bias = Parameter(torch.Tensor(n_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(self.layers[0].weight)
            bound = 1 / math.sqrt(fan_in)
            nn.init.uniform_(self.bias, -bound, bound)

    def to(self, *args, **kwargs):
        self = super().to(*args, **kwargs)
        for i in range(len(self.layers)):
            self.layers[i] = self.layers[i].to(*args, **kwargs)
        return self

    def forward(self, input):
        output = input
        for masked_linear in reversed(self.layers):
            output = masked_linear(output)
        if self.bias is not None:
            return output + self.bias
        return output


class OneHiddenLayerNet(nn.Module):
    def __init__(self, n_hidden, structure, tree_type):
        super(OneHiddenLayerNet, self).__init__()
        self.fc1_bis = None
        self.root = None
        num_factors = int(np.log2(n_hidden))
        self.n_hidden = n_hidden

        # Layers
        if structure == 'dense':
            self.fc1 = nn.Linear(n_hidden, n_hidden)
        elif structure == 'B':
            self.fc1 = NaiveButterfly(n_hidden)
        elif structure == 'BP':
            permutation = src.utils.get_permutation_matrix(num_factors, 'bit-reversal')
            self.fc1 = NaiveButterfly(n_hidden, permutation)
        elif structure == 'BPBP':
            permutation = src.utils.get_permutation_matrix(num_factors, 'bit-reversal')
            self.fc1 = NaiveButterfly(n_hidden, permutation, bias=False)
            self.fc1_bis = NaiveButterfly(n_hidden, permutation)
        self.fc2 = nn.Linear(n_hidden, 10)

        # Tree for projection if any
        if tree_type is not None:
            if tree_type == 'balanced':
                generate_tree = src.tree.generate_balanced_tree
            else:
                assert tree_type == 'comb'
                generate_tree = src.tree.generate_comb_tree
            self.root = generate_tree(0, num_factors, num_factors, use_tensor_supp=True)

    def to(self, *args, **kwargs):
        self = super().to(*args, **kwargs)
        self.fc1 = self.fc1.to(*args, **kwargs)
        if self.fc1_bis is not None:
            self.fc1_bis = self.fc1_bis.to(*args, **kwargs)
        if self.root is not None:
            self.root = self.root.to(*args, **kwargs)
        return self

    def forward(self, x):
        x = x.flatten(start_dim=1)
        n_in = x.shape[-1]
        x = F.pad(x, (0, self.n_hidden - n_in))
        x = self.fc1(x)
        if self.fc1_bis is not None:
            x = self.fc1_bis(x)
        x = F.relu(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


class TwoHiddenLayersNet(nn.Module):
    def __init__(self, n_in, n_hidden, structure, tree_type=None):
        super(TwoHiddenLayersNet, self).__init__()
        self.fc1_bis = None
        self.root = None
        num_factors = int(np.log2(n_hidden))

        # Layers
        self.fc0 = nn.Linear(n_in, n_hidden)
        if structure == 'dense':
            self.fc1 = nn.Linear(n_hidden, n_hidden)
        elif structure == 'B':
            self.fc1 = NaiveButterfly(n_hidden)
        elif structure == 'BP':
            permutation = src.utils.get_permutation_matrix(num_factors, 'bit-reversal')
            self.fc1 = NaiveButterfly(n_hidden, permutation)
        elif structure == 'BPBP':
            permutation = src.utils.get_permutation_matrix(num_factors, 'bit-reversal')
            self.fc1 = NaiveButterfly(n_hidden, permutation, bias=False)
            self.fc1_bis = NaiveButterfly(n_hidden, permutation)
        self.fc2 = nn.Linear(n_hidden, 10)

        # Tree for projection if any
        if tree_type is not None:
            if tree_type == 'balanced':
                generate_tree = src.tree.generate_balanced_tree
            else:
                assert tree_type == 'comb'
                generate_tree = src.tree.generate_comb_tree
            self.root = generate_tree(0, num_factors, num_factors, use_tensor_supp=True)

    def to(self, *args, **kwargs):
        self = super().to(*args, **kwargs)
        self.fc1 = self.fc1.to(*args, **kwargs)
        if self.fc1_bis is not None:
            self.fc1_bis = self.fc1_bis.to(*args, **kwargs)
        if self.root is not None:
            self.root = self.root.to(*args, **kwargs)
        return self

    def forward(self, x):
        x = x.flatten(start_dim=1)
        x = self.fc0(x)
        x = F.relu(x)
        x = self.fc1(x)
        if self.fc1_bis is not None:
            x = self.fc1_bis(x)
        x = F.relu(x)
        x = self.fc2(x)
        output = F.log_softmax(x, dim=1)
        return output


def train(args, model, device, train_loader, optimizer, epoch):
    losses = src.utils.AverageMeter('Loss', ':.4e')
    top1 = src.utils.AverageMeter('Acc@1', ':6.2f')
    model.train()
    end = time.time()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)

        # save metrics
        acc1 = src.utils.accuracy(output, target, topk=(1, ))
        losses.update(loss.item(), data.size(0))
        top1.update(acc1[0], data.size(0))

        loss.backward()
        optimizer.step()
        if model.root is not None:
            assert isinstance(model.fc1, nn.Linear)
            with torch.no_grad():
                model.fc1.weight.data = torch_project_p_model_p_fixed(model.fc1.weight.data, model.root)
        if batch_idx % args.log_interval == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                100. * batch_idx / len(train_loader), loss.item()))
            if args.dry_run:
                break
    total_time = time.time() - end
    return total_time, losses.avg, top1.avg.item()


def test(model, device, test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            test_loss += F.nll_loss(output, target, reduction='sum').item()  # sum up batch loss
            pred = output.argmax(dim=1, keepdim=True)  # get the index of the max log-probability
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)

    print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n'.format(
        test_loss, correct, len(test_loader.dataset),
        100. * correct / len(test_loader.dataset)))

    return test_loss, 100. * correct / len(test_loader.dataset)


def main():
    # Training settings
    parser = argparse.ArgumentParser(description='PyTorch MNIST / CIFAR-10 Example')
    parser.add_argument('--batch_size', type=int, default=64, metavar='N',
                        help='input batch size for training (default: 64)')
    parser.add_argument('--test_batch_size', type=int, default=1000, metavar='N',
                        help='input batch size for testing (default: 1000)')
    parser.add_argument('--num_workers', type=int, default=4)
    parser.add_argument('--epochs', type=int, default=14, metavar='N',
                        help='number of epochs to train (default: 14)')
    parser.add_argument('--lr', type=float, default=1.0, metavar='LR',
                        help='learning rate (default: 1.0)')
    parser.add_argument('--gamma', type=float, default=0.7, metavar='M',
                        help='Learning rate step gamma (default: 0.7)')
    parser.add_argument('--no_cuda', action='store_true', default=False,
                        help='disables CUDA training')
    parser.add_argument('--dry_run', action='store_true', default=False,
                        help='quickly check a single pass')
    parser.add_argument('--seed', type=int, default=1, metavar='S',
                        help='random seed (default: 1)')
    parser.add_argument('--log_interval', type=int, default=10, metavar='N',
                        help='how many batches to wait before logging training status')
    parser.add_argument('--save_model', action='store_true', default=False,
                        help='For Saving the current Model')
    parser.add_argument('--n_hidden', type=int, default=1024)
    parser.add_argument('--structure', choices=['dense', 'B', 'BP', 'BPBP', 'dense+balanced'], default="dense")
    parser.add_argument('--tree_type', choices=['comb', 'balanced'], default=None)
    parser.add_argument('--data_path', type=str, default='./data')
    parser.add_argument("--dataset", choices=["CIFAR10", "MNIST"], default="CIFAR10")
    parser.add_argument('--arch', choices=["1HL", "2HL"], default="1HL")
    parser.add_argument('--optimizer', choices=["SGD", "Adam", "SGDMomentum", "Adadelta"], default="Adam")
    parser.add_argument('--results_dir', type=str, default=".")
    args = parser.parse_args()

    if args.structure == 'dense+balanced':
        args.structure = 'dense'
        args.tree_type = 'balanced'

    # Stats
    results_dir = Path(args.results_dir)
    results_dir.mkdir(exist_ok=True, parents=True)
    script_params = ["batch_size", "epochs", "lr", "gamma", "n_hidden", "structure", "tree_type", "dataset",
                     "arch", "optimizer"]
    name = f"-".join([f"{param}_{getattr(args, param)}" for param in script_params])
    metrics = ["epoch", "time", "loss", "accuracy", "loss_val", "accuracy_val"]
    stats = src.utils.Stats(results_dir / f"{name}.pkl", script_params + metrics)
    print(name)

    # Preliminaries
    use_cuda = not args.no_cuda and torch.cuda.is_available()
    torch.manual_seed(args.seed)
    device = torch.device("cuda" if use_cuda else "cpu")
    train_kwargs = {'batch_size': args.batch_size}
    test_kwargs = {'batch_size': args.test_batch_size}

    # Dataset / dataloader
    if use_cuda:
        cuda_kwargs = {'num_workers': args.num_workers,
                       'pin_memory': True,
                       'shuffle': True}
        train_kwargs.update(cuda_kwargs)
        test_kwargs.update(cuda_kwargs)
    if args.dataset == "CIFAR10":
        transform = transforms.Compose(
            [transforms.Grayscale(),
             transforms.ToTensor(),
             transforms.Normalize((0.5,), (0.5,))])
        dataset1 = datasets.CIFAR10(args.data_path, train=True, download=True, transform=transform)
        dataset2 = datasets.CIFAR10(args.data_path, train=False, transform=transform)
        n_in = 3 * 32 * 32
    else:
        transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
            ])
        dataset1 = datasets.MNIST(args.data_path, train=True, download=True, transform=transform)
        dataset2 = datasets.MNIST(args.data_path, train=False, transform=transform)
        n_in = 28 * 28
    train_loader = torch.utils.data.DataLoader(dataset1, **train_kwargs)
    test_loader = torch.utils.data.DataLoader(dataset2, **test_kwargs)

    # Model architecture
    if args.arch == "1HL":
        model = OneHiddenLayerNet(args.n_hidden, args.structure, args.tree_type).to(device)
    else:
        assert args.arch == "2HL"
        model = TwoHiddenLayersNet(n_in, args.n_hidden, args.structure, args.tree_type).to(device)
    print([param.shape for param in model.parameters()])

    # Optimizer
    if args.optimizer == 'Adam':
        optimizer = optim.Adam(model.parameters(), lr=args.lr)
    elif args.optimizer == 'SGD':
        optimizer = optim.SGD(model.parameters(), lr=args.lr)
    elif args.optimizer == 'SGDMomentum':
        optimizer = optim.SGD(model.parameters(), lr=args.lr, momentum=0.9)
    else:
        assert args.optimizer == 'Adadelta'
        optimizer = optim.Adadelta(model.parameters(), lr=args.lr)
    scheduler = StepLR(optimizer, step_size=10, gamma=args.gamma)

    # Training
    for epoch in range(1, args.epochs + 1):
        time, loss, acc = train(args, model, device, train_loader, optimizer, epoch)
        loss_val, acc_val = test(model, device, test_loader)
        scheduler.step()
        stats.update([getattr(args, param) for param in script_params] + [epoch, time, loss, acc, loss_val, acc_val])
        print(stats.stats[["epoch", "time", "loss", "accuracy", "loss_val", "accuracy_val"]])

    if args.save_model:
        torch.save(model.state_dict(), "model.pt")


if __name__ == '__main__':
    main()
