import scipy.linalg
import src.tree as tree
import src.factorization as fact
import src.utils as utils
import itertools
import matplotlib.pyplot as plt
import scipy.fftpack
import numpy as np
from pathlib import Path
import argparse


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Learn fast transforms with BP model, with exhaustive search on "
                                                 "permutations.")
    parser.add_argument("--num_factors", type=int, help="Choose 2 or 3, since 4 is not tractable.")
    parser.add_argument("--matrix", type=str, choices=["dft", "hadamard", "random_butterfly"])
    parser.add_argument("--results_dir", type=str, default=".")
    args = parser.parse_args()

    if args.matrix == "dft":
        matrix = scipy.linalg.dft(2 ** args.num_factors)
    elif args.matrix == "hadamard":
        matrix = scipy.linalg.hadamard(2 ** args.num_factors)
    else:
        assert args.matrix == "random_butterfly"
        supp = utils.support_DFT(args.num_factors)
        matrix = utils.generate_random_matrix(supp)
    print(matrix.shape)

    permutations = []
    losses = []
    candidates = list(itertools.permutations(range(2 ** args.num_factors)))
    print(len(candidates))
    for i, permutation in enumerate(candidates):
        permuted_matrix = utils.permute_columns(list(permutation), matrix)
        root = tree.generate_balanced_tree(0, args.num_factors, args.num_factors)
        result = fact.tree_hierarchical_factorization(root, permuted_matrix)
        loss = utils.error_cal(result, permuted_matrix)
        permutations.append(permutation)
        losses.append(loss)
        if i % 100 == 0:
            print(f"After candidate {i} / {len(candidates)}", [permutations[i] for i in range(len(losses)) if losses[i] < 1e-5])

    all_permutations = np.array([permutations[i] for i in range(len(losses)) if losses[i] < 1e-5])

    results_dir = Path(args.results_dir)
    np.save(results_dir / "all_permutations.npy", all_permutations)

    plt.plot(losses)
    plt.ylabel("loss")
    plt.xlabel("permutation id")
    plt.title(f"learning {args.matrix}, size = 2^{args.num_factors}")
    plt.savefig(results_dir / "learn_loss.png")
