import time
import argparse
import src.utils as utils
import src.factorization as fact
import scipy.linalg
import numpy as np


def parse_args():
    parser = argparse.ArgumentParser(description="Hierarchical matrix factorization.")
    parser.add_argument("--num_factors", type=int, default=6)
    parser.add_argument("--matrix", choices=["random", "hadamard", "dft"], default="hadamard")
    parser.add_argument("--tree", type=str, choices=["comb", "balanced", "inversed_comb"], default="balanced")
    parser.add_argument("--max_depth", default=-1, type=int, help="-1 is full depth.")
    return parser.parse_args()


def main(args):
    # Generate the matrix to factorize
    if args.matrix == "random":
        print("Generating random matrix to factorize...")
        supp = utils.support_DFT(args.num_factors)
        matrix = utils.generate_random_matrix(supp)
    elif args.matrix == "dft":
        print("Generating DFT matrix to factorize...")
        matrix = scipy.linalg.dft(2 ** args.num_factors)
    else:
        assert args.matrix == "hadamard"
        print("Generating Hadamard matrix to factorize...")
        matrix = scipy.linalg.hadamard(2 ** args.num_factors)

    # Permutation
    best = np.inf

    # Hierarchical factorization
    begin = time.time()
    for perm in ["000", "001", "010", "011", "100", "101", "110", "111"]:
        p = utils.get_permutation_matrix(args.num_factors, perm)
        product, factors = fact.project_BP_model_P_fixed(matrix, args.tree, p, args.max_depth, return_factors=True)
        loss = utils.error_cal(factors + [p], matrix, relative = False)
        best = min(best, loss)
    end = time.time()

    # Loss
    loss = utils.error_cal(factors + [p], matrix)
    print("Loss:", best)
    print("Factorization time:", end - begin)


if __name__ == '__main__':
    args = parse_args()
    main(args)
