import numpy as np
from sklearn.decomposition import SparseCoder
import src.utils as utils
import src.factorization as fact
import src.archive.optimization as optim


def learning_with_butterfly_model(data, col_sparsity, num_iter, tree, perm_name, coder_method, stats, max_depth,
                                  random_permute_MEG, code_init, global_optim, lr, grad_iter, optimize):
    if random_permute_MEG:
        print("Permuting MEG rows randomly")
        data = data[np.random.permutation(len(data))]
    data_pad = utils.add_padding(data)

    if code_init == "random":
        code = np.random.randn(data_pad.shape[0], data_pad.shape[1])
    else:
        assert code_init == "eye"
        code = np.eye(data_pad.shape[0], data_pad.shape[1])
    num_factors = int(np.log2(data_pad.shape[1]))

    factors = []
    dictionary = np.zeros((2 ** num_factors, 2 ** num_factors))
    p = None
    if perm_name != "all_8":
        p = utils.get_permutation_matrix(num_factors, perm_name)

    for i in range(num_iter):
        print(f"Iteration {i}")
        print("Pseudo inverse")
        matrix = np.linalg.pinv(code) @ data_pad
        if p is None:
            print(f"BP projection, where P can be any of the fixed 8, and {tree} tree, max_depth={max_depth}")
            # matrix_product = fact.project_BP_model_8_perm_fixed(matrix, tree, max_depth)
            raise NotImplementedError
        else:
            print(f"BP projection, with fixed P = {perm_name}, and {tree} tree, max_depth={max_depth}")
            factors, product = fact.project_BP_model_P_fixed(matrix, tree, p, max_depth, return_factors=True)
        print(f"Sparse coding with {coder_method}, k={col_sparsity}")
        norms = np.linalg.norm(product, axis=1)
        indices = np.argwhere(norms > 0)
        dictionary[indices] = product[indices] / norms[indices, np.newaxis]
        coder = SparseCoder(
            dictionary=dictionary,
            transform_n_nonzero_coefs=col_sparsity,
            transform_algorithm=coder_method
        )
        code = coder.transform(data_pad)
        code[:, indices] = code[:, indices] / norms[indices]

        if global_optim:
            print("Global optimization with fixed supports")
            all_factors = [code] + factors + [p]
            gradient_method = optim.FixedSupportGradientOptimization(data_pad, [1] * (1 + len(factors)) + [0],
                                                                     *all_factors, cuda=True)
            gradient_method.train(lr=lr, optimize=optimize, num_iter=grad_iter, verbose_freq=100)
            code, *factors, _ = gradient_method.return_numpy_masked_factors()
            dictionary = utils.product_of_factors(factors) @ p

        loss = utils.relative_error_padded_product(data, code, *factors, p)
        print("Loss:", loss)
        stats.update([col_sparsity, tree, perm_name, coder_method, random_permute_MEG, code_init, global_optim, lr,
                      grad_iter, optimize, i, loss])
    return (code, *factors, p)


def method_of_optimal_directions(data, col_sparsity, num_iter, random_permute_MEG, code_init, coder_method, stats,
                                 global_optim, lr, grad_iter, optimize):
    if random_permute_MEG:
        print("Permuting MEG rows randomly")
        data = data[np.random.permutation(len(data))]
    data_pad = utils.add_padding(data)

    if code_init == "random":
        code = np.random.randn(data_pad.shape[0], data_pad.shape[1])
    else:
        assert code_init == "eye"
        code = np.eye(data_pad.shape[0], data_pad.shape[1])
    num_factors = int(np.log2(data_pad.shape[1]))

    dictionary = np.zeros((2 ** num_factors, 2 ** num_factors))
    for i in range(num_iter):
        print(f"Iteration {i}")
        print("Pseudo inverse")
        matrix = np.linalg.pinv(code) @ data_pad
        norms = np.linalg.norm(matrix, axis=1)
        indices = np.argwhere(norms > 0)
        dictionary[indices] = matrix[indices] / norms[indices, np.newaxis]
        print(f"Sparse coding with {coder_method}, k={col_sparsity}")
        coder = SparseCoder(
            dictionary=dictionary,
            transform_n_nonzero_coefs=col_sparsity,
            transform_algorithm=coder_method
        )
        code = coder.transform(data_pad)

        loss = utils.relative_error_padded_product(data, code, dictionary)
        print("Loss:", loss)
        stats.update([col_sparsity, coder_method, random_permute_MEG, code_init, global_optim, lr,
                      grad_iter, optimize, i, loss])
    return code, dictionary
