import src.utils as utils
import numpy as np
from sklearn.decomposition import SparseCoder
from pathlib import Path
import inspect
import src.factorization as fact
import src.archive.optimization as optim


class DictionaryLearningMethod:
    def __init__(self, data, results_dir, seed, **params):
        self.data = data
        self.data_pad = utils.add_padding(self.data)

        results_dir = Path(results_dir)
        results_dir.mkdir(exist_ok=True, parents=True)
        self.save_name = "-".join([f"{key}_{value}" for key, value in params.items()])
        self.stats = utils.Stats(
            results_dir / f"{self.save_name}.pkl",
            ["seed"] + [key for key in params.keys()] + ["iter", "loss"]
        )
        np.random.seed(seed)

    def random_permute_data_rows(self):
        self.data = self.data[np.random.permutation(len(self.data))]
        self.data_pad = utils.add_padding(self.data)


class MethodOfOptimalDirections(DictionaryLearningMethod):
    def __init__(self, data, results_dir, seed, code_init, num_iter, coder_method, code_sparsity, use_padding,
                 bp_tree=None, bp_depth=None, bp_perm=None,
                 optim_iter=None, optim_freq=None, optim_lr=None, optimizer=None, cuda=0):
        local_dict = locals()
        signature = list(inspect.signature(MethodOfOptimalDirections.__init__).parameters)[3:]
        self.params = {param: local_dict[param] for param in signature}
        DictionaryLearningMethod.__init__(self, data, results_dir, **self.params)
        self.num_iter = num_iter
        self.coder_method = coder_method
        self.code_sparsity = code_sparsity
        # BP parameters
        self.bp_tree = bp_tree
        self.bp_depth = bp_depth
        self.bp_perm = bp_perm
        # Global optimization parameters
        self.optim_iter = optim_iter
        self.optim_freq = optim_freq
        self.optim_lr = optim_lr
        self.optimizer = optimizer
        self.cuda = bool(cuda)
        # Which matrix to factorize
        self.use_padding = use_padding
        self.data_matrix = self.data_pad if use_padding else data
        print("Learn dictionary of a matrix of size ", self.data_matrix.shape)
        if code_init == "random":
            self.code = np.random.randn(self.data_matrix.shape[0], self.data_matrix.shape[1])
        else:
            assert code_init == "eye"
            self.code = np.eye(self.data_matrix.shape[0], self.data_matrix.shape[1])

    def factorize(self):
        # Initialization
        code = self.code
        dictionary = np.zeros((self.data_matrix.shape[1], self.data_matrix.shape[1]))

        # BP projection
        perm = None
        root = None
        bp_projection = self.bp_tree is not None and self.bp_depth is not None and self.bp_perm is not None
        if bp_projection:
            perm = utils.get_permutation_matrix(int(np.log2(self.data_matrix.shape[1])), self.bp_perm)

        # Global optimization
        factors = []
        global_optim = bp_projection and self.optim_lr is not None and self.optim_iter is not None and \
                       self.optim_freq is not None and self.optimizer is not None

        # Dictionary learning
        for i in range(1, self.num_iter + 1):
            print(f"Iteration {i}")
            print("Pseudo inverse")
            matrix = np.linalg.pinv(code) @ self.data_matrix
            if self.use_padding:
                matrix[:, 204:255] = 0
                if bp_projection:
                    print(f"BP projection, with {self.bp_tree} tree, {self.bp_perm} permutation, {self.bp_depth} max depth")
                    matrix, factors, root = fact.project_BP_model_P_fixed(matrix, self.bp_tree, perm, self.bp_depth,
                                                                          return_factors=True, return_root=True)
            norms = np.linalg.norm(matrix, axis=1)
            indices = np.argwhere(norms > 0)
            dictionary[indices] = matrix[indices] / norms[indices, np.newaxis]
            print(f"Sparse coding with {self.coder_method}, k={self.code_sparsity}")
            coder = SparseCoder(
                dictionary=dictionary,
                transform_n_nonzero_coefs=self.code_sparsity,
                transform_algorithm=self.coder_method
            )
            code = coder.transform(self.data_matrix)

            if global_optim and i % self.optim_freq == 0:
                code[:, indices] = code[:, indices] / norms[indices]
                print(f"Global optimization with fixed supports, with {self.optimizer}, lr={self.optim_lr}, "
                      f"{self.optim_iter} iterations. {'Use cuda' if self.cuda else 'Do not use cuda'}")
                all_factors = [code] + factors + [perm]
                if self.bp_depth == 1:
                    all_supports = [code, root.left.support, root.right.support, perm]
                elif self.bp_depth == 2:
                    all_supports = [code, root.left.support, root.right.left.support, root.right.right.support, perm]
                else:
                    # TODO in order search of supports of root
                    raise NotImplementedError
                gradient_method = optim.FixedSupportGradientOptimization(self.data_matrix, [1] * (1 + len(factors)) + [0],
                                                                         all_factors, all_supports, cuda=self.cuda)
                gradient_method.train(lr=self.optim_lr, optimize=self.optimizer, num_iter=self.optim_iter,
                                      verbose_freq=20)
                code, *factors, _ = gradient_method.return_numpy_masked_factors()
                dictionary = utils.product_of_factors(factors) @ perm

            loss = utils.relative_error_padded_product(self.data, code, dictionary)
            print("Loss:", loss)
            stats_row = self.params.copy()
            stats_row.update({"loss": loss, "iter": i})
            self.stats.update_with_dict(stats_row)

        return code, dictionary


class PalmDictionaryLearning(DictionaryLearningMethod):
    pass
    # TODO