import torch
import torch.nn as nn
import numpy as np
import torch.optim as optim


class FixedSupportGradientOptimization(object):
    def __init__(self, matrix, requires_grad_list, all_factors, all_supports, cuda=False):
        self.matrix = torch.from_numpy(matrix).to(torch.double)
        self.factors = [torch.from_numpy(factor).to(torch.double) for factor in all_factors]
        self.supports = [torch.from_numpy(supp).to(torch.bool) for supp in all_supports]

        # Cuda
        if cuda:
            self.matrix = self.matrix.cuda()
            self.factors = [X.cuda() for X in self.factors]
            self.supports = [supp.cuda() for supp in self.supports]

        # Requires grad
        for i in range(len(self.factors)):
            if requires_grad_list[i]:
                self.factors[i] = self.factors[i].requires_grad_()

    def masked_factors(self):
        return [X * supp for (X, supp) in zip(self.factors, self.supports)]

    def train(self, lr, num_iter=1000, optimize='SGD', verbose_freq=-1):
        loss_fn = nn.MSELoss(reduction='sum')
        re_array = []
        best = np.inf
        matrix_norm = torch.norm(self.matrix)

        # Choose optimizer
        if optimize == 'Adam':
            optimizer = optim.Adam(self.factors, lr=lr)
        elif optimize == 'SGDMomentum':
            optimizer = optim.SGD(self.factors, lr=lr, momentum=0.9)
        else:
            assert optimize == 'SGD'
            optimizer = optim.SGD(self.factors, lr=lr)

        for index in range(num_iter):
            optimizer.zero_grad()
            masked_factors = self.masked_factors()
            approx_mat = masked_factors[0]
            for i in range(1, len(masked_factors)):
                approx_mat = torch.mm(approx_mat, masked_factors[i])
            loss = loss_fn(approx_mat, self.matrix)
            loss.backward()
            optimizer.step()

            with torch.no_grad():
                relative_error = torch.div(torch.sqrt(loss), matrix_norm).item()
                re_array.append(relative_error)
                if relative_error < best:
                    best = relative_error
                if verbose_freq and index % verbose_freq == 0:
                    print(f"Iteration {index}: {relative_error}")
        return re_array, best

    def return_numpy_masked_factors(self):
        return tuple([factor.detach().cpu().numpy() for factor in self.masked_factors()])
